/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "config.h"

static char const rcsid[] UNUSED =
"$Id: log.c,v 1.45 2008/07/02 11:02:10 jcpp Exp $";



#include <sys/types.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <unistd.h>
#include <fcntl.h>
#include <syslog.h>
#include <utime.h>
#include <errno.h>

#include <stdarg.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "lmtp.h"

#define BUFLEN 8192

int logrotate(struct fstate *state) {
  int fd, fdtmp;
  char tmpname[PATH_MAX];
  char domain[255];
  FILE *tmp;
  FILE *msg;
  int c;
  char *tmplog;

  getuserdomain(state->qrecipient, domain);

  if ((msg=opentemplate(state->log_msg, domain))==NULL)
    return -1;

  if ((fd=open(state->logname, O_RDONLY|O_EXCL|O_NONBLOCK)) < 0) {
    fclose(msg);
    return -1;
  }

  if ((tmp=newmessage(state, "log", tmpname))==NULL) {
    fclose(msg);
    close(fd);
    return -1;
  }

  while((c=fgetc(msg))!=EOF) {
    if (c=='%') {
      c=fgetc(msg);
      switch(c) {
      case 'D':
        fputs(domain, tmp);
        break;
      case 'R':
        fputs(state->qrecipient, tmp);
        break;
      case 's': {
	char stamp[100];
	strftime(stamp, sizeof(stamp)-1,
		 "%Y%m%d%H%M%S", localtime(&(state->date)));
	fputs(stamp, tmp);
      }
      break;
      case 'L': {
	char buf[BUFLEN];
	ssize_t len;
	while((len=read(fd, buf, BUFLEN))>0) {
	  fwrite(buf, 1, len, tmp);
	}
	if (len <0) {
	  fclose(msg);
	  fclose(tmp);
	  close(fd);
	  unlink(tmpname);
	  return -1;
	}
      }
      break;
      case EOF:
        break;
      default:
        fputc(c, tmp);
      }
    } else {
      fputc(c, tmp);
    }
  }

  fclose(msg);

  if (endmessage(state, "log", tmpname, tmp)==EOF) {
    close(fd);
    return -1;
  }

  /* we use a secondary file to handle correctly accessed time  */
  /* and rotate log file considering age of this secondary file */
  tmplog = strdup(state->logname);
  tmplog = strcat(tmplog, ".tmp");
  unlink(tmplog);
  fdtmp = open(tmplog, O_WRONLY|O_TRUNC|O_CREAT|O_NONBLOCK);
  close(fdtmp);

  close(fd);
  fd = open(state->logname, O_WRONLY|O_TRUNC|O_EXCL|O_NONBLOCK);
  if (state->filesystem_with_noatime) {
    int saved_errno = errno;
    struct utimbuf times;
    times.actime = state->date;
    times.modtime = state->date; 
    utime(state->logname, &times);
    errno = saved_errno;
  }
  return fd;
}

int logopen(struct fstate *state) {
  struct stat sb;
  struct stat sbtmp;
  int fd, fdtmp;
  char *tmplog;

  if (state->log_msg && state->log_msg[0] && 
      state->log_mintime>0 && state->log_maxsize>0 &&
      state->logname) {
    if (stat(state->logname, &sb)==-1) {
      return -1;
    }
    /* we use a secondary file to handle correctly accessed time  */
    /* and rotate log file considering age of this secondary file */
    tmplog = strdup(state->logname);
    tmplog = strcat(tmplog, ".tmp");
    if (stat(tmplog, &sbtmp)==-1) {
      /* create secondary file if inexistant */
      fdtmp = open(tmplog, O_WRONLY|O_TRUNC|O_CREAT|O_NONBLOCK);
      close(fdtmp);
      stat(tmplog, &sbtmp);
    }
    if (state->date-sbtmp.st_atime > state->log_mintime &&
	sb.st_size > (off_t)state->log_maxsize) {
      if ((fd = logrotate(state))>0) {
	syslog(LOG_INFO, "log rotation message send to <%s> (size=%lu)",
	       state->qrecipient, (unsigned long)(sb.st_size));
	  return fd;
      }
    }
  }

  return open(state->logname, O_WRONLY|O_APPEND|O_EXCL|O_NONBLOCK);
}

static void loghdr(FILE *f, struct headers *hdr) {
  char *s = hdr->header;
  while(*s) {
    if (isprint((int)*s) || *s == '\n' || *s == '\t') {
      fputc(*s, f);
    } else {
      fputc('.', f);
    }
    s++;
  }
  fputc('\n', f);
}

#ifdef UNUSED_ACTION_LOG
/* specific action log. I've no use of this for now, so I will not
   release it. If somebody want, let me know */

/*
.IP "\fBaction.log.deliver.format\fP format"
.IP "\fBaction.log.forward.format\fP format"
.IP "\fBaction.log.bounce.format\fP format"
.IP "\fBaction.log.drop.format\fP format"
.IP "\fBaction.log.error.format\fP format"

If set, send an extra syslog message when an 'deliver', 'forward',
'bounce', 'drop' or 'error' event occurs in delivery.

Within the format string, two letter sequences begining by '%'
character are expanded like this:
.RS
.IP \fB%m\fP
the mailbox
.IP \fP%a\fP
the action string
.IP \fP%s\fP
the message size
.IP \fP%r\fP
the recipient
.IP \fP%f\fP
the sender of the message
.RE

.IP "\fBaction.log.deliver.facility\fP facility"
.IP "\fBaction.log.forward.facility\fP facility"
.IP "\fBaction.log.bounce.facility\fP facility"
.IP "\fBaction.log.drop.facility\fP facility"
.IP "\fBaction.log.error.facility\fP facility"

Set the syslog facility for extra action logs. If unset, the
value is taken from \fBsyslog.facility\fP setting.
*/

static void log_action(struct fstate *state, struct actionq *act, char *mbox,
		     int level, int facility, char *format) {
  if (format && *format && (level & act->logit)) {
    char buf[LINELEN];
    strfmt(buf, sizeof(buf), format, "$ms$a$r$f",
	   mbox, (int)state->msgsize, act->action,
	   state->qrecipient, state->sender);
    if (facility == 0) 
      facility = state->syslog_facility;
    syslog(facility | LOG_INFO, "%s", buf);
  }
}
#endif

int logit(struct fstate *state, struct actionq *act, 
	  unsigned char type, char *fmt, ...) {
  va_list ap;
  char *mbox;
  double xdelay;
  struct timeval t;

  va_start(ap, fmt);
  vsnprintf(act->action, ERRMSGMAX, fmt, ap);
  va_end(ap);
  act->action[ERRMSGMAX]=0;
  { 
    char *s=act->action;
    while(*s) {
      if (*s=='\r' || *s=='\n')
	*s=' ';
      s++;
    }
  }

  gettimeofday(&t, NULL);
  xdelay = (double)(t.tv_sec - state->date_deliver.tv_sec) + 
    (double)(t.tv_usec - state->date_deliver.tv_usec) / 1000000;

  act->logit = ((act->logmask & type) != 0);

  mbox = quote_addr821(state->rcpt->mbox);
  
  if (!mbox || strcmp(mbox, state->qrecipient)==0)
    syslog(LOG_NOTICE, ID_FMT "to=<%s>, delay=%.2f, action=%s", 
	   state->counter, state->qrecipient, xdelay, act->action);
  else
    syslog(LOG_NOTICE, ID_FMT "to=<%s>, mbox='%s', delay=%.2f, action=%s", 
	   state->counter,  state->qrecipient, 
	   mbox, xdelay, act->action);

  if (mbox && state->syslog_host && act->logit)
    my_syslog(state, "%s: from=<%s>, size=%lu, action=%s",
	      mbox, state->sender, (unsigned long)state->msgsize, act->action);

#ifdef UNUSED_ACTION_LOG
  log_action(state, act, mbox, ACTION_DELIVER,
	     state->action_log_deliver_facility,
	     state->action_log_deliver_format);
  log_action(state, act, mbox, ACTION_FORWARD,
	     state->action_log_forward_facility,
	     state->action_log_forward_format);
  log_action(state, act, mbox, ACTION_BOUNCE,
	     state->action_log_bounce_facility,
	     state->action_log_bounce_format);
  log_action(state, act, mbox, ACTION_DROP,
	     state->action_log_drop_facility,
	     state->action_log_drop_format);
  log_action(state, act, mbox, ACTION_ERROR,
	     state->action_log_error_facility,
	     state->action_log_error_format);
#endif

  free(mbox);

  return 0;
}

int logflush(struct fstate *state, struct actionq *act) { 
  int fd;
  FILE *f;
  char bufdate[100];
  struct headers *hl=state->hdrlist;
  double xdelay;
  struct timeval t;
  struct stat saved_file_stat;

  if (state->logname==NULL || state->logname[0] == 0)
     return 0;

  if ((fd=logopen(state))==-1)
    return 0;

  if ((f=fdopen(fd, "a"))==NULL) {
    close(fd);
    return -1;
  }

  if (state->filesystem_change_atime_when_write) /* workaround netapp bug */
    fstat(fd, &saved_file_stat);

  gettimeofday(&t, NULL);
  xdelay = (double)(t.tv_sec - state->date_deliver.tv_sec) + 
    (double)(t.tv_usec - state->date_deliver.tv_usec) / 1000000;

  strftime(bufdate, sizeof(bufdate)-1, STRFTIME_FMT,
	   usertime(state));

  fprintf(f, "From %s\t%s\n",
	  state->sender[0] ? state->sender : "MAILER-DAEMON" , bufdate);
  fprintf(f, "Status: score=" SCORE_FORMAT
	  ", size=%lu, lines=%lu, rcpts=%lu, delay=%.2fs",
	  state->score, (unsigned long)state->msgsize,
	  (unsigned long)state->nblines, (unsigned long)state->nrcpts_r,
	  xdelay);
  if (state->av.nvirus > 0 && state->av.virus[0]) {
    fprintf(f, ", virus=%s", state->av.virus);
  }
  fprintf(f, "\n");
  if (state->trace) {
    fprintf(f, "Trace: %s\n", state->trace);
  }
  act = state->actions;
  
  while(act) {
    if (act->action[0])
      fprintf(f, "Action: %s\n", act->action);
    act = act->next;
  }
  while(hl) {
    if (hl->trace & TRACE_ON)
      loghdr(f, hl);
    hl=hl->next;
  }
  fputc('\n', f);
  fclose(f);
  if (state->filesystem_change_atime_when_write) {
    struct utimbuf times;
    times.actime = saved_file_stat.st_atime;
    times.modtime = state->date; /* should to another stat ?? */
    utime(state->logname, &times);
  }
  return 0;
}

