/*
   Copyright 2005 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "config.h"

static char const rcsid[] UNUSED =
"$Id: sieve.c,v 1.10 2005/09/12 08:59:53 lwa Exp $";


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>

#include "lmtp.h"
#include "sieve/sieve.h"

static int valid_folder(void *message, const char *folder) {
  return 1;
}


static unsigned msg_spam(void *message) {
  struct fstate *state = message;
  spam_assassin(state, NULL);
  if (state->spamassassin_done < 0)
    return 0; /* message was not tested for spam */
  if (state->spamassassin_is_spam)
    return 10; /* message was tested and definitely contains spam */
  if (state->spamassassin_score <= 1)
    return 1; /* message was tested and is clear of spam */
  if (state->spamassassin_score >= 9)
    return 9;
  return state->spamassassin_score;
}

static unsigned msg_virus(void *message) {
  struct fstate *state = message;
  av_scan(state);
  if (!state->av.scanned) 
    return 0; /* message was not tested for viruses */
  if (state->av.nvirus <= 0) 
    return 1; /* message was tested and contains no known viruses */
  return 5;   /* message was tested and definitely contains a known virus */
}

static char *msg_from(void *message) {
  struct fstate *state = message;
  return state->sender;
}
static char *msg_to(void *message) {
  struct fstate *state = message;
  return state->qrecipient;
}
static size_t msg_size(void *message) {
  struct fstate *state = message;
  return state->msgsize;
}
static char *msg_header(void *message, const char *header,
			int decoded, void **hstate) {
  struct fstate *state = message;
  struct headers **hlp, *hl;
  size_t hlen = strlen(header);
  if (hstate && *hstate) 
    hlp = *hstate;
  else
    hlp = &state->hdrlist;
  /*  fprintf(stderr, "msg_header(%s) %p %p %p\n", header, hstate, state->hdrlist, hl); */
  if (*hlp == NULL)
    return NULL;
  hl = *hlp;
  while(hl) {
    /*                    fprintf(stderr, "check header %s in %s (%d %d)\n", header, hl->header, hlen, hl->dd_len);    */
    if (hlen == hl->dd_len && strncasecmp(hl->header, header, hlen) == 0) {
      if (hstate) {
	*hstate = &hl->next;
      }
      /*      fprintf(stderr, "found, state=%p, header=%s\n", *hstate,  hl->header + hl->dd_len); */
      return (decoded ? utf8_header(hl) : hl->header) + hl->dd_len+1;
    }
    hl = hl->next;
  }
  return NULL;
}

int filtersieve(struct fstate *state, char *file) {
  FILE *ffilter;
  int final = FILTER_CONTINUE;
  if (file!= NULL && (ffilter=fopen(file, "r"))!=NULL) {
    struct sieve_methods Pmethods = {0,};
    struct sieve_result Presult = {0,};
    int ret;
    char **l;
    int drop ;
    Pmethods.valid_folder = valid_folder;
    Pmethods.separator = '+';
    Pmethods.from = msg_from;
    Pmethods.to = msg_to;
    Pmethods.header = msg_header;
    Pmethods.size = msg_size;
    Pmethods.spam = msg_spam;
    Pmethods.virus= msg_virus;
    ret = sieve((sieve_fgetc_t)fgetc, ffilter, &Pmethods, state, &Presult);
    
    fclose(ffilter);

    drop = !Presult.implicit_keep;

    if (ret != 0) {
      push_action_log(state, "Sieve error near line %d: %s",
		      Presult.lastline, Presult.error);
      sieve_result_free(&Presult);
      return FILTER_STOP;
    }

    l = Presult.fileinto;
    if (l) {
      while(*l) {
	drop = 0;
	push_action(state, *l, action_deliver0);
	l++;
      }
    }

    l = Presult.redirect;
    if (l) {
      while(*l) {
	drop = 0;
	if (**l) push_action(state, *l, action_forward2);
	l++;
      }
    }

    if (Presult.reject) {
      char *end = strpbrk(Presult.reject, "\r\n");
      if (end)
	*end = 0;
      push_action(state, Presult.reject, action_bounce);
      drop = 0;
    }
    if (drop) {
      push_action(state, "", action_drop); /* for logging */
    }

    sieve_result_free(&Presult);

    return Presult.implicit_keep ? FILTER_ENDFILE : FILTER_STOP;
  }
  return final;
}

